''' Pesquisas em grafos

    Exemplo:
    >>> from iaed.graph import bfs, dfs
    >>> gr3 = { 'A': ['B', 'C'], 'B': ['E'], 'C': ['B', 'F'],
        'D': ['C', 'G'], 'E': ['F'], 'F': [], 'G': ['F'] }
    >>> print(bfs(gr3, 'A'))
    >>> print(dfs(gr3))
'''
def bfs(graph: dict, start: str) -> dict:
    ''' Travessia em largura num grafo '''
    out = { start: [0, None] } # node: [ distance, predecessor ]
    queue = (start,)
    while queue:
        node = queue[0]
        queue = queue[1:]
        for dest in graph[node]:
            if dest not in out: # is white
                # is gray
                out[dest] = [out[node][0]+1, node]
                queue += (dest,)
        # is black
    return out


def dfs(graph: dict) -> dict:
    ''' Travessia em profundidade num grafo '''
    def visit(graph, node, out, time, pred = ''):
        ''' Visita um nó do grafo em profundidade '''
        out[node] = [ time, -1, pred] # is gray
        time += 1
        for dest in graph[node]:
            if dest not in out: # is white
                time = visit(graph, dest, out, time, node)
        out[node][1] = time
        return time +1
    out = {} # node: [ discovery, finish, predecessor ]
    time = 1
    for node in graph:
        if node not in out: # is white
            time = visit(graph, node, out, time)
    return out

if __name__ == '__main__':
    from sys import argv
    from ast import literal_eval
    if len(argv) > 1:
        for s in argv[1:]:
            gr = literal_eval(s)
            st = list(gr)
            st.sort()
            print(bfs(gr, st[0]))
            print(dfs(gr))
        exit(0)
    # dictionary of adjacency lists(ordered)
    gr1 = { 'A': ['B', 'C', 'D'], 'B': ['E'], 'C': ['B', 'G'],
        'D': ['C', 'G'], 'E': ['C', 'F'], 'F': ['C', 'H'],
        'G': ['F', 'H', 'I'], 'H': ['E', 'I'], 'I': ['F'] }
    print(bfs(gr1, 'A'))
    print(dfs(gr1))
    gr2 = { 'u': ['v', 'x'], 'v': ['y'], 'w': ['y', 'z'],
        'x': ['v'], 'y': ['x'], 'z': ['z'] }
    print(bfs(gr2, 'u'))
    print(dfs(gr2))
    gr3 = { 'A': ['B', 'C'], 'B': ['E'], 'C': ['B', 'F'],
        'D': ['C', 'G'], 'E': ['F'], 'F': [], 'G': ['F'] }
    print(bfs(gr3, 'A'))
    print(dfs(gr3))



''' Operações em árvores de pesquisa binárias (BST) [não equilibradas]

    Exemplo:
    >>> from iaed.btree import Btree
    >>> x = Btree()
    >>> for i in (12, 6, 8, 16, 24, 7, 15, 2, 21):
    >>>     x.insert(i)
    >>> x.preorder()
    >>> x.posorder()
    >>> x.order()
'''

class _node:
    ''' Nó da árvore '''
    def __init__(self, value, left=None, right=None):
        ''' Cria um nó '''
        self._value = value
        self._left = left
        self._right = right
    def value(self, val = ...):
        ''' Valor do nó '''
        if val != ...:
            self._value = val
        return self._value
    def left(self, val = ...):
        ''' Sub-nó esquerdo '''
        if val != ...:
            self._left = val
        return self._left
    def right(self, val = ...):
        ''' Sub-nó direito '''
        if val != ...:
            self._right = val
        return self._right

class Btree:
    ''' Árvore binária '''
    def __init__(self, keyf = lambda x: x):
        ''' keyf: função que devolve a chave de um nó '''
        self._tree = None
        self._keyf = keyf

    def insert(self, value):
        ''' Operação de inserção '''
        def _insert(node, keyf, value):
            ''' Insere um nó numa subárvore '''
            if not node: # new node
                return _node(value)
            cmp = keyf(value) - keyf(node.value())
            if not cmp: # replace node
                return _node(value)
            if cmp < 0: # go left
                node.left(_insert(node.left(), keyf, value))
            else: # go right
                node.right(_insert(node.right(), keyf, value))
            return node
        self._tree = _insert(self._tree, self._keyf, value)

    def search(self, key):
        ''' Operação de procura '''
        def _search(node, keyf, key):
            ''' Procura um nó numa subárvore '''
            if not node:
                return None
            cmp = key - keyf(node.value())
            if not cmp:
                return node.value()
            if cmp < 0:
                return _search(node.left(), keyf, key)
            return _search(node.right(), keyf, key)
        return _search(self._tree, self._keyf, key)

    def max(self):
        ''' Devolve o valor do maior nó da árvore '''
        node = self._tree
        if not node:
            return None
        while node.right():
            node = node.right()
        return node.value()

    def min(self):
        ''' Devolve o valor do menor nó da árvore '''
        node = self._tree
        if not node:
            return None
        while node.left():
            node = node.left()
        return node.value()

    def remove(self, key):
        ''' Operação de remoção '''
        def _remove(node, keyf, key):
            ''' Remove um nó de uma subárvore '''
            if not node:
                return None
            cmp = key - keyf(node.value())
            if cmp < 0:
                node.left(_remove(node.left(), keyf, key))
            elif cmp > 0:
                node.right(_remove(node.right(), keyf, key))
            else:
                if node.left() and node.right(): # caso 3
                    aux = node.left() # find highest value
                    while aux.right():
                        aux = aux.right()
                    aux._value, node._value = node.value(), aux.value()
                    node.left(_remove(node.left(), keyf, aux.value()))
                else:
                    if not node.left() and not node.right():
                        node = None
                    elif not node.left():
                        node = node.right()
                    else:
                        node = node.left()
            return node
        self._tree = _remove(self._tree, self._keyf, key)

    def count(self):
        ''' Devolve o número de nós numa árvore '''
        def _count(node):
            ''' Devolve o número de subnós de um nó '''
            if not node:
                return 0
            return 1 + _count(node.left()) + _count(node.right())
        return _count(self._tree)

    def height(self):
        ''' Devolve a altura máxima de uma árvore '''
        def _height(node):
            ''' Devolve a altura máxima de um nó '''
            if not node:
                return 0 # -1 for height(leaf)==0
            hleft = _height(node.left())
            hright = _height(node.right())
            return hleft+1 if hleft > hright else hright+1
        return _height(self._tree)

    def preorder(self, visit = print):
        ''' Efetua uma travessia pré-order '''
        def _preorder(node, visit):
            ''' Visita um nó em pré-order '''
            if node:
                visit(node.value())
                _preorder(node.left(), visit)
                _preorder(node.right(), visit)
        _preorder(self._tree, visit)

    def posorder(self, visit = print):
        ''' Efetua uma travessia pós-order '''
        def _posorder(node, visit):
            ''' Visita um nó em pós-order '''
            if node:
                _posorder(node.left(), visit)
                _posorder(node.right(), visit)
                visit(node.value())
        _posorder(self._tree, visit)

    def order(self, visit = print):
        ''' Efetua uma travessia in-order '''
        def _inorder(node, visit):
            ''' Visita um nó em in-order '''
            if node:
                _inorder(node.left(), visit)
                visit(node.value())
                _inorder(node.right(), visit)
        _inorder(self._tree, visit)

if __name__ == '__main__':
    x = Btree()
    x.insert(12)
    x.insert(6)
    x.insert(8)
    x.insert(16)
    x.insert(24)
    from io import StringIO
    import sys
    out = sys.stdout
    sys.stdout = StringIO()
    assert x.count() == 5, "count test"
    assert x.height() == 3, "height test"
    assert x.min() == 6, "min test"
    assert x.max() == 24, "max test"
    assert x.search(8) == 8, "search test"
    assert x.search(7) is None, "fail search test"
    x.preorder()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "12\n6\n8\n16\n24\n", "preorder test"
    sys.stdout = StringIO()
    x.posorder()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "8\n6\n24\n16\n12\n", "posorder test"
    sys.stdout = StringIO()
    x.order()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "6\n8\n12\n16\n24\n", "inorder test"
    sys.stdout = StringIO()
    x.remove(12)
    x.preorder()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "8\n6\n16\n24\n", "remove root test"
    sys.stdout = StringIO()
    x.remove(16)
    x.preorder()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "8\n6\n24\n", "remove right test"
    sys.stdout = StringIO()
    x.insert(123)
    x.insert(16)
    x.insert(12)
    x.remove(16)
    x.preorder()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "8\n6\n24\n12\n123\n", "remove left test"
    sys.stdout = StringIO()
    x.remove(123)
    x.preorder()
    sys.stdout.seek(0)
    assert sys.stdout.read() == "8\n6\n24\n12\n", "remove left test"
    sys.stdout = out
    print("OK")



''' Tabelas de dispersão por dispersão dupla (double-hashing)

    Exemplo para uma tabela de dimensão 11 sem redimensionamento (dens=1):
    >>> from iaed.hashdbl import DblHash
    >>> x = DblHash(size=11,dens=1)
    >>> for n in (15,14,12,24,18,11,21,48,32):
    >>>     x.insert(n)
    >>> print(x)
'''

class DblHash:
    ''' Dispersão por dispersão dupla (double-hashing) '''
    class Del:
        ''' deleted value '''
    def __init__(self, keyf = lambda x: x, dbl = lambda x: 3*x+1, dens = .5,
        hashf = hash, size = 199):
        ''' keyf: função que devolve o valor de um elemento
            dbl: função de hashing secundária
            dens: densidade limite para redimensionamento
            hashf: função de hashing primária
            size: dimensão inicial da tabela '''
        self._keyf = keyf
        self._dbl = dbl
        self._hashf = hashf
        self._size = size
        self._dens = dens
        self._tab = [None] * size
        self._busy = 0

    def _expand(self):
        ''' Redimensiona a tabela '''
        tab = self._tab
        self._size = self._size * 2 + 1
        self._tab = [None] * self._size
        self._busy = 0
        for val in tab:
            if val is not None:
                self.insert(val)

    def insert(self, value):
        ''' Insere um elemento '''
        i = self._hashf(self._keyf(value)) % self._size
        k = self._dbl(self._keyf(value))
        while self._tab[i] is not None and self._tab[i] != self.Del:
            i = (i + k) % self._size
        self._tab[i] = value
        self._busy += 1
        if self._busy / self._size > self._dens:
            self._expand()

    def remove(self, value):
        ''' Remove um elemento '''
        i = self._hashf(self._keyf(value)) % self._size
        k = self._dbl(self._keyf(value))
        while self._tab[i] is not None:
            if self._tab[i] != self.Del and \
                self._keyf(self._tab[i]) == self._keyf(value):
                break
            i = (i + k) % self._size
        if self._tab[i] is not None:
            self._tab[i] = self.Del
            self._busy -= 1

    def search(self, key):
        ''' Procura um elemento '''
        i = self._hashf(key) % self._size
        k = self._dbl(key)
        while self._tab[i] is not None:
            if self._tab[i] != self.Del and self._keyf(self._tab[i]) == key:
                return self._tab[i]
            i = (i + k) % self._size
        return None

    def __repr__(self):
        ''' Devolve a representação interna da tabela '''
        out = ''
        for i, val in enumerate(self._tab):
            if val is None:
                continue
            out += str(i) + ': ' + str(val) + '\n'
        return out + str(self._size) + ', ' + str(self._busy) + ', ' + str(self._dens)

    def __getitem__(self, index):
        ''' Devolve o n-ésimo elemento da tabela '''
        return self._tab[index] if 0 <= index < self._size else None

if __name__ == '__main__':
    x = DblHash(size=11,dens=1)
    for n in (15,14,12,24,18,11,21,48,32):
        x.insert(n)
    print(x)



''' Algoritmos de pesquisa

    Exemplo:
    >>> from iaed import search
    >>> from random import randint
    >>> l = [ randint(0, 100) for _ in range (20) ]
    >>> search.search(l, 12)
    >>> l.sort()
    >>> search.sorted_search(l, 12)
    >>> search.binary_search(l, 12)
'''
from random import randint
from time import time

def binary_search(array, value, l=0, r=None) : # array must be sorted
    ''' Pesquisa binária em vetor ordenado '''
    if r is None:
        r = len(array) # return index or None
    print(l, r)
    if r <= l :
        return None
    mid = l + (r - l) // 2
    if array[mid] == value :
        return mid
    if array[mid] > value :
        return binary_search(array, value, l, mid-1)
    return binary_search(array, value, mid+1, r)

def sorted_search(array, value) : # array must be sorted
    ''' Pesquisa linear ordenada '''
    for pos, val in enumerate(array) :
        if val == value :
            return pos
        if val > value :
            return None # give up
    return None

def search(array, value) :
    ''' Pesquisa linear '''
    for pos, val in enumerate(array) :
        if val == value :
            return pos
    return None

def mklistset(length=10000, high=100000):
    ''' Cria um vetor de 'length' inteiros entre 0 e 'high' '''
    s = set([randint(0, high) for _ in range(length)])
    return s, list(s)

def search_set(s, vec):
    ''' Procura num conjunto 's' cada um dos elementos em 'vec'.
        Retorna o tempo gasto (seg) e o número de elementos encontrados.'''
    count = 0
    start = time()
    for x in range(vec):
        if x in s:
            count += 1
    return time - start, count

def search_list(l, vec):
    ''' Procura numa lista 'l' cada um dos elementos em 'vec'.
        Retorna o tempo gasto (seg) e o número de elementos encontrados.'''
    count = 0
    start = time()
    for x in range(vec):
        if x in l:
            count += 1
    return time - start, count

def search_test(length=10000, count=10000):
    ''' Compara tempos de pesquisa em lista e set. '''
    l, s = mklistset(length)
    _, vec = mklistset(count)
    print("set", search_set(s, vec))
    print("list", search_list(l, vec))



#! /usr/bin/env python3
''' Algoritmos de ordenação


    >>> from iaed import sorting
    >>> v = [34, 3, 36, 11, 8, 4, 22, 20, 12]
    >>> sorting.dbg = True # ver cada iterada
    >>> sorting.selection(v)

    Em algoritmos recursivos pode ser necessário aumentar o limite de recursão:
    >>> from sys import setrecursionlimit
    >>> setrecursionlimit(50000)
'''
from random import randint
from timeit import timeit

dbg = False # debug
''' para cada iterada colocar a True '''

def run(algorithm, array, pkg = 'iaed.sorting') :
    ''' Execução temporizada
        >>> from iaed import sorting
        >>> sorting.dbg = True # ver cada iterada
        >>> sorting.run("bubble", sorting.mkarray(12,100))
    '''
    return timeit(f"{algorithm}({array})",f"from {pkg} import {algorithm}", number=1)

# from iaed.sorting import bubble
# timeit(f"{algorithm}({array})",f"from __main__ import {algorithm}", number=1)

def mkarray(length, high=1000, low=0) :
    ''' Cria vetor com valores inteiros aleatórios '''
    return [ randint(low, high) for _ in range(length) ]

lower = 'abcdefghijklmnopqrstuvwxyz'
upper = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
digit = '0123456789'
hexdigit = '0123456789ABCDEF'
charset = lower + upper + digit
charspc = charset + ' '
def rndstr(length, chars = charset):
    ''' Cria uma cadeia de carateres com o comprimento dado (length)
        constituída por carateres do conjunto dado (chars) '''
    out = ''
    for _ in range(length):
        out += chars[randint(0,len(chars)-1)]
    if out[0] == ' ' or out[-1] == ' ':
        return rndstr(length, chars)
    return out

def mkstrvec(length, width, chars = charset):
    ''' Cria um vetor de cadeias de carateres '''
    return [ rndstr(width, chars) for _ in range(length) ]

def bubble0(array) :
    ''' Ordenação por borbulhamento (esq -> dir)'''
    n = len(array)
    for i in range(n) :
        for j in range(n - i - 1) : # left to right
            if array[j] > array[j+1] :
                array[j], array[j+1] = array[j+1], array[j]
        if dbg:
            print(array)
    return array

def bubble1(array) :
    ''' Ordenação por borbulhamento invertido (dir -> esq)'''
    n = len(array) -1
    for i in range(n) :
        for j in range(n, i, -1) : # right to left
            if array[j] < array[j-1] :
                array[j], array[j-1] = array[j-1], array[j]
        if dbg:
            print(array)
    return array

def bubble(array) :
    ''' Ordenação por borbulhamento ótimizado (esq -> dir)'''
    n = len(array)
    for i in range(n) :
        is_sorted = True
        for j in range(n - i - 1) :
            if array[j] > array[j+1] :
                array[j], array[j+1] = array[j+1], array[j]
                is_sorted = False
        if is_sorted :
            break
        if dbg:
            print(array)
    return array

def insertion(array) :
    ''' Ordenação por inserção '''
    for i in range(1, len(array)) :
        item = array[i]
        j = i - 1
        while j >= 0 and array[j] > item :
            array[j + 1] = array[j]
            j -= 1
        array[j + 1] = item
        if dbg:
            print(array)
    return array

# selection([15,14,12,24,18,11,21,48,32])
def selection(array) :
    ''' Ordenação por seleção '''
    for i in range(len(array)) :
        pos = i
        for j in range(i+1, len(array)) :
            if array[pos] > array[j] :
                pos = j
        array[pos], array[i] = array[i], array[pos]
        if dbg:
            print(array)
    return array

def shell(array) :
    ''' Ordenação Shell '''
    h = 1
    while h <= len(array) // 9 :
        h = 3 * h + 1

    while h > 0 :
        for i in range(h, len(array)) :
            j = i
            item = array[i]
            while j >= h and item <= array[j - h]:
                array[j] = array[j - h]
                j -= h
            array[j] = item
        h //= 3
        if dbg:
            print(array)
    return array

def partition(a, l, r) :
    ''' Partição de um vetor no QuickSort '''
    i, j, pivot = l, r -1, a[r]
    while i <= j :
        while a[i] <= pivot and i <= j :
            i += 1
        while pivot <= a[j] and j >= i :
            j -= 1
        if i < j :
            a[i], a[j] = a[j], a[i]
    a[i], a[r] = a[r], a[i]
    return i

def quick(array, l=0, r=None) :
    ''' Ordenação rápida (Quick)'''
    if r is None :
        r = len(array) -1
    if r <= l :
        return array
    # if r-l <= M : return insertion(a, l, r)
    i = partition(array, l, r)
    quick(array, l, i-1)
    quick(array, i+1, r)
    return array

def merge_sorted(a, b) :
    ''' Fusão de dois vetores ordenados '''
    if len(a) == 0 :
        return b
    if len(b) == 0 :
        return a
    ret = []
    i = j = 0
    while len(ret) < len(a) + len(b) :
        if a[i] <= b[j] :
            ret += [ a[i] ]
            i += 1
            if i == len(a) :
                break
        else :
            ret += [ b[j] ]
            j += 1
            if j == len(b) :
                break
    return ret + a[i:] + b[j:]

def merge(array) :
    ''' Ordenação por fusão '''
    if len(array) < 2 :
        return array
    mid = len(array) // 2
    return merge_sorted(merge(array[:mid]), merge(array[mid:]))

def parent(k):
    ''' Índice do elemento hierarquicamente anterior (pai) '''
    return (k+1)//2-1
def left(k):
    ''' Índice do subelemento hieraequicamente à esquerda (filho esquerdo) '''
    return 2*k+1
def right(k):
    ''' Índice do subelemento hieraequicamente à direita (filho direita) '''
    return 2*(k+1)
def fix_down(a, l, r, k) :
    ''' Recolocação descendente do elemento índice k '''
    largest, ileft, iright = k, left(k), right(k)
    if ileft <= r and a[largest] < a[ileft] :
        largest = l + left(k-l)
    if iright <= r and a[largest] < a[iright] :
        largest = l + right(k-l)
    if largest != k :
        a[k], a[largest] = a[largest], a[k]
        fix_down(a, l, r, largest)
def fix_up(vec, k):
    ''' Recolocação ascendente do elemento índice k '''
    while k > 0 and vec[parent(k)] < vec[k]:
        vec[k], vec[parent(k)] = vec[parent(k)], vec[k]
        k = parent(k)
def insert(vec, item):
    ''' Insere um elemento num amontoado '''
    vec += [item]
    fix_up(vec, len(vec)-1)
def remove(vec):
    ''' Remove um elemento de um amontoado '''
    item, vec[0], vec[-1] = vec[0],  vec[-1], vec[0]
    fix_down(vec, 0, len(vec)-2, 0)
    del vec[-1]
    return item
def buildheap(a, l=0, r=None) :
    ''' Constrói um amontoado '''
    if r is None :
        r = len(a)-1
    heapsize = r-l+1
    k = heapsize//2-1
    while k >= l :
        fix_down(a, l, r, l+k)
        k -= 1
def heapsort(a, l=0, r=None) :
    ''' Ordenação por amontoado '''
    if r is None :
        r = len(a)-1
    buildheap(a, l, r)
    if dbg:
        print('build',a)
    while r - l > 0 :
        a[l], a[r] = a[r], a[l]
        r -= 1
        fix_down(a, l, r, l)
        if dbg:
            print(a)
    return a

def distcount(a, M=1000) :
    ''' counting sort with M keys in [0,999] '''
    cnt = [0] * (M+1)
    b = [0] * len(a)
    for i in range(len(a)) :
        cnt[a[i]+1] += 1
    for j in range(M) :
        cnt[j+1] += cnt[j]
    for i in range(len(a)) :
        b[cnt[a[i]]] = a[i]
        cnt[a[i]] += 1
    return b

def distcount2(a, M=1000) :
    ''' backward counting sort with M keys in [0,999] '''
    cnt = [0] * M
    b = [0] * len(a)
    for i in range(len(a)) :
        cnt[a[i]] += 1
    for j in range(M-1) :
        cnt[j+1] += cnt[j]
    for i in range(len(a)) :
        cnt[a[i]] -= 1
        b[cnt[a[i]]] = a[i]
    return b

def radixLSD(a, M, bytesword, digit) :
    ''' Ordenação radix Least Significant Digit (LSD) first

    sort a vector of two (2) decimal (10) digits:
    radixLSD([15,14,12,24,18,11,21,48,32],10,2,lambda x,y: int(str(x)[y]))

    sort a vector of integers upto 6-bits with 2-bits (4 possible values) at a time
    (3-sets of 2-bits is a 6-bit number: between 0 and 63)
    radixLSD(v, 4, 3, lambda x,y: int(('0'*6+bin(x)[2:])[-6:][(y*2:(1+y)*2],2))
    '''
    while bytesword > 0 :
        bytesword -= 1
        count = [0] * (M + 1)
        aux = [0] * len(a)
        for i in range(len(a)) :
            count[digit(a[i], bytesword) + 1] += 1
        for j in range(M) :
            count[j+1] += count[j]
        for i in range(len(a)) :
            aux[count[digit(a[i], bytesword)]] = a[i]
            count[digit(a[i], bytesword)] += 1
        a = list(aux)
        if dbg:
            print(a)
    return a

def quicksortBin(a, l, r, w, bitsword) :
    ''' Binary Quick sort '''
    def digit(a,b) :
        return (a >> (bitsword-b)) & 1
    if r <= l or w > bitsword :
        return None
    i, j = l, r
    while j != i :
        while digit(a[i], w) == 0 and i < j :
            i += 1
        while digit(a[j], w) == 1 and j > i :
            j -= 1
        a[i], a[j] = a[j], a[i]
    if digit(a[r], w) == 0 :
        j += 1
    quicksortBin(a, l, j-1, w+1, bitsword)
    quicksortBin(a, j, r, w+1, bitsword)
    return a

def quicksortBinv(a, l, r, w, bitsword) :
    ''' Binary Quick sort inverted '''
    def digit(a,b) :
        return (a >> b) & 1
    if r <= l or w < 0 :
        return None
    i, j = l, r
    while j != i :
        while digit(a[i], w) == 0 and i < j :
            i += 1
        while digit(a[j], w) == 1 and j > i :
            j -= 1
        a[i], a[j] = a[j], a[i]
    if digit(a[r], w) == 0 :
        j += 1
    quicksortBinv(a, l, j-1, w-1, bitsword)
    quicksortBinv(a, j, r, w-1, bitsword)
    return a

def insertion_sort(array, l=0, r=None):
    ''' Insertion sort for radix MSD '''
    if r is None:
        r = len(array) - 1
    for i in range(l + 1, r + 1):
        item = array[i]
        j = i - 1
        while j >= l and array[j] > item:
            array[j + 1] = array[j]
            j -= 1
        array[j + 1] = item
    return array

def radixMSD(a, l, r, w, bytesword, digit, M, ins=32) :
    ''' Ordenação radix Most Significant Digit (MSD) first

    sort a vector of decimal values (M=10) with up to three (bytesword=3) digits:
    v = [198, 981, 136, 448, 148, 445, 827, 442, 133]
    radixMSD(v, 0, 8, 0, 3, lambda x, y: int(str(x)[y]), 10, ins=0)
    '''
    def bin(x): return l+count[x]
    if r <= l: return a
    if dbg: print('radix:', w, 'digit from', l, 'to', r, '=', a[l:r+1])
    oif w > bytesword :
        return a
    if r-l <= ins :
        insertion_sort(a, l, r)
        return a
    count = [0] * (M + 1)
    aux = [0] * (r -l + 1)
    for i in range(len(aux)) :
        count[digit(a[l+i], w) + 1] += 1
    for j in range(M) :
        count[j+1] += count[j]
    for i in range(len(aux)) :
        aux[count[digit(a[l+i], w)]] = a[l+i]
        count[digit(a[l+i], w)] += 1
    for i in range(len(aux)) :
        a[l+i] = aux[i]
    radixMSD(a, l, bin(0)-1, w+1, bytesword, digit, M, ins)
    for j in range(M-1) :
        radixMSD(a, bin(j), bin(j+1)-1, w+1, bytesword, digit, M, ins)
    return a

# insertion_sort array slices and merge_sorted them
# https://github.com/python/cpython/blob/main/Objects/listobject.c #2191
# min_run: https://hg.python.org/cpython/file/tip/Objects/listsort.txt
def timsort(array): # O(n log2 n)
    ''' Ordenação timsort '''
    min_run = 32
    n = len(array)
    for i in range(0, n, min_run): # sort each sliced sub-array
        insertion_sort(array, i, min((i + min_run - 1), n - 1))
    size = min_run
    while size < n: # merge from min_run to n
        for start in range(0, n, size * 2):
            midpoint = start + size - 1
            end = min((start + size * 2 - 1), (n-1))
            merged_array = merge_sorted(
                array[start:midpoint + 1],
                array[midpoint + 1:end + 1])
            array[start:start + len(merged_array)] = merged_array
        size *= 2 # double size at each iteration
    return array

if __name__ == "__main__" :
    from sys import argv, setrecursionlimit
    from time import time
    setrecursionlimit(50000)
    if len(argv) == 1 :
        print('USAGE:', argv[0], ' count [algorithm/value]')
        print('\tSorting: bubble0, bubble1, bubble, insertion, selection,')
        print('\t\tshell, quick, merge, heapsort, distcount, distcount2')
    if len(argv) == 2 :
        v = mkarray(int(argv[1]))
        t = time()
        bubble0(v)
        print('bubble0:', time()-t)
        t = time()
        bubble1(v)
        print('bubble1:', time()-t)
        t = time()
        bubble(v)
        print('bubble:', time()-t)
        t = time()
        insertion(v)
        print('insertion:', time()-t)
        t = time()
        selection(v)
        print('selection:', time()-t)
        t = time()
        shell(v)
        print('shell:', time()-t)
        t = time()
        quick(v)
        print('quick:', time()-t)
        t = time()
        merge(v)
        print('merge:', time()-t)
        t = time()
        heapsort(v)
        print('heap:', time()-t)
        t = time()
        timsort(v)
        print('tim:', time()-t)
        t = time()
        v.sort()
        print('sort:', time()-t)
    if len(argv) == 3 :
        if argv[2].isnumeric() :
            from iaed.search import search, sorted_search, binary_search
            v = mkarray(int(argv[1]))
            v.sort()
            t = time()
            p = search(v, int(argv[2]))
            print('search: #', p, time()-t)
            t = time()
            p = sorted_search(v, int(argv[2]))
            print('sorted: #', p, time()-t)
            t = time()
            p = binary_search(v, int(argv[2]))
            print('binary: #', p, time()-t)
        else :
            f = locals()[argv[2]]
            t = time()
            f(mkarray(int(argv[1])))
            print(argv[2] + ':', time()-t)



# @(c) Pedro Reis dos Santos, IST Ulisboa, 2026
"""
Introdução ao algoritmos e estruturas de dados

Algoritmos de pesquisa:
>>> help("iaed.search")

Algoritmos de ordenação:
>>> help("iaed.sorting")

Estruturas de dados: tabelas de dispersão (hashext, hashlin, hashdbl),
    árvores(btree, AVL), grafos(graph)
>>> help("iaed.hashext")
"""

__author__ = 'Pedro Reis dos Santos <reis.santos@tecnico.ulisboa.pt>'
__web__ = 'https://github.com/pedroreissantos/iaed.git'
__version__ = '1.0.3'
__all__ = ['search', 'sorting', 'hashext', 'hashlin', 'hashdbl',
        'btree', 'avl', 'graph', 'fib'] # visible modules within the package



''' Termos da sequência de Fibonacci

    >>> from iaed import fib
    >>> fib.recurs(35)
    >>> fib.iter(35)
'''
from time import time

def recurs(n):
    ''' Cálculo recursivo do N-ésimo termo da sequência de Fibonacci '''
    if n < 2:
        return n
    return recurs(n - 2) + recurs(n - 1)

def iter(n):
    ''' Cálculo iterativo do N-ésimo termo da sequência de Fibonacci '''
    a, b = 0, 1
    while n > 1:
        a, b, n = b, a+b, n-1
    return b

def test(n):
    start = time()
    print("recurs", recurs(n), "em", time()-start, "seg")
    start = time()
    print("iter", iter(n), "em", time()-start, "seg")



#!/usr/bin/env python3
"""IAED: Introduction to algorithms and data structures
search: linear and binary
sorting: selection, insertion, bubble, shell, quick, merge, heap, counting, radix LSB/MSB and priority queues
hash-tables: external, linear probing, double hashing
tree: binary search, AVL
graph: breadth transversal, depth transversal
"""

import setuptools

with open("README.md", encoding="utf8") as fh:
    readme = fh.read()

setuptools.setup(name='iaed',
	version='1.0.3', # 1.0.3 test
	author='Pedro Reis dos Santos',
	author_email="reis.santos@tecnico.ulisboa.pt",
	description="IAED: Introduction to algorithms and data structures",
	long_description=readme,
	long_description_content_type="text/x-rst",
	license = 'MIT',
	url="https://github.com/pedroreissantos/iaed",
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		'Intended Audience :: Developers',
		'Topic :: Software Development :: Build Tools',
		'Development Status :: 4 - Beta',
		'Environment :: Console',
	],
	python_requires='>=3.6',
	py_modules = ['iaed'],
	packages=setuptools.find_packages(),
)



''' Operações em árvores equilibradas AVL (Adelson-Velsky and Landis)

    Exemplo:
    >>> from iaed.avl import AVL
    >>> t = AVL()
    >>> for i in (15, 14, 12, 24, 18, 11, 21, 48, 32):
    >>>     t.insert(i)
    >>> t.preorder()
    >>> t.remove(14)
    >>> t.posorder()
'''

from iaed.btree import _node, Btree

class _hnode(_node):
    ''' Nó da árvore com pesos '''
    def __init__(self, value, height=1, left=None, right=None):
        ''' Cria um nó '''
        super().__init__(value, left, right)
        self._height = height
    def height(self, val = ...):
        ''' Peso do nó '''
        if val != ...:
            self._height = val
        return self._height

class AVL(Btree):
    ''' Balanceamento AVL de árvores (Btree) '''
    def __init__(self, keyf = lambda x: x):
        ''' keyf: função que devolve a chave de um nó '''
        super().__init__(keyf)

    @staticmethod
    def _balance(node):
        ''' Balanceamento de um nó '''
        def height(node):
            ''' Devolve a altura de um nó '''
            return node.height() if node else 0

        def update_height(node):
            ''' Atualiza a altura de nó '''
            left = height(node.left())
            right = height(node.right())
            node.height(left+1 if left > right else right+1)

        def rot_l(node):
            ''' Efetua uma rotação simples à esquerda '''
            print('rotL', node.value())
            aux = node.right()
            node.right(aux.left())
            aux.left(node)
            update_height(node)
            update_height(aux)
            return aux

        def rot_r(node):
            ''' Efetua uma rotação simples à direita '''
            print('rotR', node.value())
            aux = node.left()
            node.left(aux.right())
            aux.right(node)
            update_height(node)
            update_height(aux)
            return aux

        def rot_lr(node):
            ''' Efetua uma rotação dupla esquerda-direita '''
            print('rotLR', node.value())
            if not node:
                return None
            node.left(rot_l(node.left()))
            return rot_r(node)

        def rot_rl(node):
            ''' Efetua uma rotação dupla direita-esquerda '''
            print('rotRL', node.value())
            if not node:
                return None
            node.right(rot_r(node.right()))
            return rot_l(node)

        def balance_factor(node):
            ''' Devolve o desiqulíbrio de um nó '''
            if not node:
                return 0
            return height(node.left()) - height(node.right())

        if not node:
            return None
        factor = balance_factor(node)
        if factor > 1:
            if balance_factor(node.left()) >= 0:
                node = rot_r(node)
            else:
                node = rot_lr(node)
        elif factor < -1:
            if balance_factor(node.right()) <= 0:
                node = rot_l(node)
            else:
                node = rot_rl(node)
        else:
            update_height(node)
        return node

    def insert(self, value):
        ''' Operação de inserção '''

        def _insert(node, keyf, value):
            ''' Insere o nó com balanceamento '''
            if not node: # new node
                return _hnode(value)
            cmp = keyf(value) - keyf(node.value())
            if not cmp: # replace node
                return _hnode(value)
            if cmp < 0: # go left
                node.left(_insert(node.left(), keyf, value))
            else: # go right
                node.right(_insert(node.right(), keyf, value))
            return AVL._balance(node)
        self._tree = _insert(self._tree, self._keyf, value)

    def remove(self, key):
        ''' Operação de remoção '''
        def _remove(node, keyf, key):
            ''' Remove o nó com balanceamento '''
            if not node:
                return None
            cmp = key - keyf(node.value())
            if cmp < 0:
                node.left(_remove(node.left(), keyf, key))
            elif cmp > 0:
                node.right(_remove(node.right(), keyf, key))
            else:
                if node.left() and node.right(): # caso 3
                    aux = node.left() # find highest value
                    while aux.right():
                        aux = aux.right()
                    aux._value, node._value = node.value(), aux.value()
                    node.left(_remove(node.left(), keyf, aux.value()))
                else:
                    if not node.left() and not node.right():
                        node = None
                    elif not node.left():
                        node = node.right()
                    else:
                        node = node.left()
            return AVL._balance(node)
        self._tree = _remove(self._tree, self._keyf, key)

if __name__ == '__main__':
    t=AVL()
    for i in (15,14,12,24,18,11,21,48,32):
        t.insert(i)
    t.preorder()
    t.remove(14)
    t.posorder()
